@@ -1,9 +1,9 @@
-from . import core, models, overwrite, plotting, satlas2_parser
+from . import core, interface, models, overwrite, plotting
 from .core import *
 from .models import *
 from .overwrite import *
 from .plotting import *
-from .satlas2_parser import *
+from .interface import *
 
 try:
     from ._version import version as __version__
