@@ -6,11 +6,12 @@ Implementation of the base Fitter, Source, Model and Parameter classes
 from __future__ import annotations
 
 import copy
-from typing import Union
+from typing import Tuple, Union
 
 import lmfit as lm
 import numdifftools as nd
 import numpy as np
+import pandas as pd
 import scipy.optimize as optimize
 from numpy.typing import ArrayLike
 
@@ -37,7 +38,7 @@ class Fitter:
         self.bounds = optimize.Bounds([], [])
         self.share = []
         self.shareModel = []
-        self.priors = []
+        self.priors = {}
         self.expressions = {}
         self.mode = 'source'
 
@@ -132,7 +133,23 @@ class Fitter:
         uncertainty : float
             Standard deviation associated with the value.
         """
-        self.priors.append((source, model, parameter_name, value, uncertainty))
+        self.priors['___'.join([source, model,
+                                parameter_name])] = (value, uncertainty)
+
+    def removeParamPrior(self, source: str, model: str,
+                         parameter_name: str) -> None:
+        """Removes a prior set on a parameter.
+
+        Parameters
+        ----------
+        source : str
+            Name of the datasource in which the parameter is present.
+        model : str
+            Name of the model in which the parameter is present.
+        parameter_name : str
+            Name of the parameter.
+        """
+        del self.priors['___'.join([source, model, parameter_name])]
 
     def addSource(self, source: 'Source') -> None:
         """Add a datasource to the Fitter structure
@@ -198,7 +215,8 @@ class Fitter:
         self.lmpars = lmpars
 
     def f(self) -> ArrayLike:
-        """Calculate the response of the models in the different sources, stacked horizontally.
+        """:meta private:
+        Calculate the response of the models in the different sources, stacked horizontally.
 
         Returns
         -------
@@ -208,7 +226,8 @@ class Fitter:
         return np.hstack([source.f() for _, source in self.sources])
 
     def y(self) -> ArrayLike:
-        """Stack the data in the different sources, horizontally.
+        """:meta private:
+        Stack the data in the different sources, horizontally.
 
         Returns
         -------
@@ -218,7 +237,8 @@ class Fitter:
         return np.hstack([source.y for _, source in self.sources])
 
     def yerr(self) -> ArrayLike:
-        """Stack the uncertainty in the different sources, horizontally.
+        """:meta private:
+        Stack the uncertainty in the different sources, horizontally.
 
         Returns
         -------
@@ -228,7 +248,8 @@ class Fitter:
         return np.hstack([source.yerr() for _, source in self.sources])
 
     def setParameters(self, params: lm.Parameters) -> None:
-        """Set the parameters of the underlying Models
+        """:meta private:
+        Set the parameters of the underlying Models
         based on a large Parameters object
 
         Parameters
@@ -242,7 +263,8 @@ class Fitter:
                     parameter_name].value = params[p].value
 
     def setUncertainties(self, params: lm.Parameters) -> None:
-        """Set the uncertainties of the underlying Models
+        """:meta private:
+        Set the uncertainties of the underlying Models
         based on a large Parameters object
 
         Parameters
@@ -255,7 +277,8 @@ class Fitter:
                 p].stderr
 
     def setCorrelations(self, params: lm.Parameters) -> None:
-        """Set the correlations of the underlying Models
+        """:meta private:
+        Set the correlations of the underlying Models
         based on a large Parameters object
 
         Parameters
@@ -281,7 +304,8 @@ class Fitter:
                 pass
 
     def resid(self) -> ArrayLike:
-        """Calculates the residuals for use in a Gaussian fitting.
+        """:meta private:
+        Calculates the residuals for use in a Gaussian fitting.
         Based on the value of :attr:`Fitter.mode`, a different method is
         used. If :attr:`Fitter.mode` is 'source', the result of :func:`~Fitter.yerr` is used.
         If :attr:`Fitter.mode` is 'combined', the denominator is calculated as
@@ -304,21 +328,26 @@ class Fitter:
         return resid
 
     def gaussianPriorResid(self) -> ArrayLike:
-        """Calculates the residual (x-xtrue)/sigma for use
+        """:meta private:
+        Calculates the residual (x-xtrue)/sigma for use
         in a Gaussian prior. The parameters for which this calculates
-        the priors are given by :func:`~Fitter.setPrior`.
+        the priors are given by :func:`~Fitter.setParamPrior`.
 
         Returns
         -------
         ArrayLike
         """
-        return np.array([
-            (self.pars[source][model][parameter].value - value) / uncertainty
-            for source, model, parameter, value, uncertainty in self.priors
-        ])
+        returnval = []
+        for key in self.priors.keys():
+            source, model, parameter = key.split('___')
+            lit, unc = self.priors[key]
+            returnval.append(
+                (self.pars[source][model][parameter].value - lit) / unc)
+        return np.array(returnval)
 
     def residualCalculation(self) -> ArrayLike:
-        """Calculates the full residual, based on :func:`~Fitter.resid`
+        """:meta private:
+        Calculates the full residual, based on :func:`~Fitter.resid`
         and :func:`~Fitter.gaussianPriorResid`
 
         Returns
@@ -328,7 +357,8 @@ class Fitter:
         return np.hstack([self.resid(), self.gaussianPriorResid()])
 
     def gaussLlh(self) -> ArrayLike:
-        """Calculate the Gaussian likelihood
+        """:meta private:
+        Calculate the Gaussian likelihood
 
         Returns
         -------
@@ -338,7 +368,8 @@ class Fitter:
         return -0.5 * resid * resid  # Faster than **2
 
     def poissonLlh(self) -> ArrayLike:
-        """Calculate the Poisson likelihood
+        """:meta private:
+        Calculate the Poisson likelihood
 
         Returns
         -------
@@ -347,13 +378,18 @@ class Fitter:
         model_calcs = self.f()
         returnvalue = self.temp_y * np.log(model_calcs) - model_calcs
         returnvalue[model_calcs <= 0] = -np.inf
+        priors = self.gaussianPriorResid()
+        if len(priors) > 1:
+            priors = -0.5 * priors * priors
+            returnvalue = np.append(returnvalue, priors)
         return returnvalue
 
     def llh(self,
             params: lm.Parameters,
             method: str = 'gaussian',
             emcee: bool = False) -> ArrayLike:
-        """Calculate the likelihood, based on the parameters and method.
+        """:meta private:
+        Calculate the likelihood, based on the parameters and method.
         In case the minimizer uses the emcee package, the array is summed to a single number.
         In case the minimizer uses any other routine, the array is multiplied by -1 to
         obtain the negative likelihood.
@@ -383,8 +419,9 @@ class Fitter:
             returnvalue = np.sum(returnvalue)
         return returnvalue
 
-    def reduction(self, r: ArrayLike) -> float:
-        """Reduces the likelihood to a single number. Used by lmfit.
+    def reductionSum(self, r: ArrayLike) -> float:
+        """:meta private:
+        Reduces the likelihood to a single number. Used by lmfit.
 
         Parameters
         ----------
@@ -398,8 +435,25 @@ class Fitter:
         """
         return np.sum(r)
 
+    def reductionSSum(self, r: ArrayLike) -> float:
+        """:meta private:
+        Reduces the likelihood to a single number. Used by lmfit.
+
+        Parameters
+        ----------
+        r : ArrayLike
+            Array of residuals
+
+        Returns
+        -------
+        float
+            Sum of array of residuals
+        """
+        return np.sum(r * r)
+
     def chisquare(self, params: lm.Parameters) -> ArrayLike:
-        """Chisquare optimization function for lmfit.
+        """:meta private:
+        Chisquare optimization function for lmfit.
 
         Parameters
         ----------
@@ -415,40 +469,20 @@ class Fitter:
         return self.residualCalculation()
 
     def _prepareFit(self):
+        """:meta private:"""
         self._createParameters()
         self._createLmParameters()
 
-    def reportFit(self,
-                  modelpars: Union[lm.Parameters, None] = None,
-                  show_correl: bool = False,
-                  min_correl: float = 0.1,
-                  sort_pars: Union[bool, callable] = False) -> str:
-        """Generate a report of the fitting results.
-
-        The report contains the best-fit values for the parameters and their uncertainties and correlations.
-
-        Parameters
-        ----------
-        modelpars : lmfit.Parameters, optional
-            Known Model Parameters
-        show_correl : bool, optional
-            Whether to show a list of sorted correlations, by default False
-        min_correl : float, optional
-            Smallest correlation in absolute value to show, by default 0.1
-        sort_pars : bool or callable, optional
-            Whether to show parameter names sorted in alphanumerical order.
-            If False (default), then the parameters will be listed in the
-            order they were added to the Parameters dictionary. If callable,
-            then this (one argument) function is used to extract a comparison
-            key from each list element.
-
-        Returns
-        -------
-        str
-            Multi-line text of fit report.
+    def revertFit(self):
+        """Reverts the parameter values to the original values.
         """
-        return lm.fit_report(self.result, modelpars, show_correl, min_correl,
-                             sort_pars)
+        params = self.result.init_values
+        for p in params.keys():
+            source_name, model_name, parameter_name = p.split('___')
+            self.pars[source_name][model_name][parameter_name].value = params[
+                p]
+        self._prepareFit()
+        self.setParameters(self.lmpars)
 
     def fit(self,
             llh: bool = False,
@@ -459,7 +493,8 @@ class Fitter:
             filename: str = None,
             steps: int = 1000,
             nwalkers: int = 50,
-            scale_covar: bool = True) -> None:
+            scale_covar: bool = True,
+            iter_cb: callable = None) -> None:
         """Perform a fit of the models (added to the sources) to the data in the sources.
         Models in the same source are summed together, models in different sources can be
         linked through their parameters.
@@ -495,6 +530,8 @@ class Fitter:
 
         kws = {}
         kwargs = {}
+        kwargs['iter_cb'] = iter_cb
+        reduce_fcn = self.reductionSum
         if llh or method.lower() == 'emcee':
             llh = True
             func = self.llh
@@ -503,6 +540,7 @@ class Fitter:
                 method = 'slsqp'
         else:
             func = self.chisquare
+            reduce_fcn = self.reductionSSum
 
         if method == 'emcee':
             llh = True
@@ -527,7 +565,6 @@ class Fitter:
         if llh:
             scale_covar = False
 
-        reduce_fcn = self.reduction
         self.result = minimize(func,
                                self.lmpars,
                                method=method,
@@ -538,18 +575,83 @@ class Fitter:
         del self.temp_y
         self.updateInfo()
 
-    def revertFit(self):
-        """Reverts the parameter values to the original values.
-        """
-        params = self.result.init_values
-        for p in params.keys():
-            source_name, model_name, parameter_name = p.split('___')
-            self.pars[source_name][model_name][parameter_name].value = params[
-                p]
-        self._prepareFit()
-        self.setParameters(self.lmpars)
+    def reportFit(self,
+                  modelpars: Union[lm.Parameters, None] = None,
+                  show_correl: bool = False,
+                  min_correl: float = 0.1,
+                  sort_pars: Union[bool, callable] = False) -> str:
+        """Generate a report of the fitting results.
 
-    def readWalk(self, filename: str):
+        The report contains the best-fit values for the parameters and their uncertainties and correlations.
+
+        Parameters
+        ----------
+        modelpars : lmfit.Parameters, optional
+            Known Model Parameters
+        show_correl : bool, optional
+            Whether to show a list of sorted correlations, by default False
+        min_correl : float, optional
+            Smallest correlation in absolute value to show, by default 0.1
+        sort_pars : bool or callable, optional
+            Whether to show parameter names sorted in alphanumerical order.
+            If False (default), then the parameters will be listed in the
+            order they were added to the Parameters dictionary. If callable,
+            then this (one argument) function is used to extract a comparison
+            key from each list element.
+
+        Returns
+        -------
+        str
+            Multi-line text of fit report.
+        """
+        return lm.fit_report(self.result, modelpars, show_correl, min_correl,
+                             sort_pars)
+
+    def createResultDataframe(self) -> pd.DataFrame:
+        """Generates a dataframe containing all information about the parameters
+        after a fit.
+
+        Returns
+        -------
+        pd.DataFrame"""
+        data = [[
+            p.split('___')[0],
+            p.split('___')[1],
+            p.split('___')[2], self.result.params[p].value,
+            self.result.params[p].stderr, self.result.params[p].min,
+            self.result.params[p].max, self.result.params[p].expr,
+            self.result.params[p].vary
+        ] for p in self.result.params]
+        columns = [
+            'Source', 'Model', 'Parameter', 'Value', 'Stderr',
+            'Boundary (min)', 'Boundary (max)', 'Expression', 'Vary'
+        ]
+        df = pd.DataFrame(data=data, columns=columns)
+        return df
+
+    def createMetadataDataframe(self) -> pd.DataFrame:
+        """Generates a dataframe containing the fitting information and
+        statistics.
+
+        Returns
+        -------
+        pd.DataFrame"""
+        columns = [
+            'Source', 'Fitting method', 'Message', 'Function evaluations',
+            'Data points', 'Variables', 'Chi-square', 'Reduced chi-square',
+            'Akaike info criterium', 'Bayesian info criterium'
+        ]
+        source = [name for (name, s) in self.sources]
+        source = ', '.join(source)
+        data = [[
+            source, self.result.method, self.result.message, self.result.nfev,
+            self.result.ndata, self.result.nvarys, self.result.chisqr,
+            self.result.redchi, self.result.aic, self.result.bic
+        ]]
+        df = pd.DataFrame(data=data, columns=columns)
+        return df
+
+    def readWalk(self, filename: str, burnin: int = 0):
         """Read and process the h5 file containing the results of a random walk.
         The parameter values and uncertainties are extracted from the walk.
 
@@ -560,7 +662,7 @@ class Fitter:
         """
         reader = SATLASHDFBackend(filename)
         # var_names = list(reader.labels)
-        data = reader.get_chain(flat=False)
+        data = reader.get_chain(flat=False, discard=burnin)
         try:
             self.result = SATLASMinimizer(self.llh, self.lmpars).process_walk(
                 self.lmpars, data)
@@ -600,6 +702,87 @@ class Fitter:
             except:
                 pass
 
+    def evaluateOverWalk(self,
+                         filename: str,
+                         burnin: int = 0,
+                         x: ArrayLike = None,
+                         evals: int = 0) -> Tuple[list, list]:
+        """The parameters saved in the h5 file are evaluated
+        in the models a specific number of times. From these evaluations, the
+        16, 50 and 84 percentiles (corresponding to the 1-sigma band) are calculated.
+
+        Parameters
+        ----------
+        filename : str
+            Filename of the random walk results.
+        burnin : int, optional
+            Amount of steps to skip, by default 0
+        x : ArrayLike, optional
+            Evaluation points for the model, defaults to the datapoints in Source
+        evals : int, optional
+            Number of selected parameter values, defaults to using all values
+
+        Returns
+        -------
+        Tuple of lists
+            A tuple with, as the first element, a list of arrays x-values for
+            which the band has been evaluated. Each source contributes an array.
+            The second element is a list of 2D arrays, one for each source. The
+            first row is the sigma- boundary, the second row is the median value
+            and the third row is the sigm+ boundary.
+        """
+        reader = SATLASHDFBackend(filename)
+        var_names = list(reader.labels)
+        data = reader.get_chain(flat=False, discard=burnin)
+        flatchain = data.reshape((-1, len(var_names)))
+        if x is None:
+            method = 'f'
+            args = ()
+        else:
+            method = 'evaluate'
+            args = (x, )
+        if evals > 0:
+            if evals < flatchain.shape[0]:
+                choices = np.random.choice(flatchain.shape[0], evals)
+                flatchain = flatchain[choices]
+        else:
+            pass
+        try:
+            names = [p for p in self.lmpars.keys()]
+        except:
+            self._prepareFit()
+            names = [p for p in self.lmpars.keys()]
+        common = [(i, name.split('___')) for i, name in enumerate(var_names)
+                  if name in names]
+        bands = []
+        X = []
+        for sample in flatchain:
+            for column, splitname in common:
+                source, model, parameter = splitname
+                self.pars[source][model][parameter].value = sample[column]
+            for i, (_, source) in enumerate(self.sources):
+                try:
+                    bands[i] = np.vstack(
+                        [bands[i], getattr(source, method)(*args)])
+                except Exception as e:
+                    bands.append(getattr(source, method)(*args))
+                if len(args) > 0:
+                    X.append(args[0])
+                else:
+                    X.append(source.x)
+        for i, band in enumerate(bands):
+            q = np.percentile(band, [16, 84], axis=0)
+            bands[i] = q
+        median = np.percentile(flatchain, 50, axis=0)
+        for column, splitname in common:
+            source, model, parameter = splitname
+            self.pars[source][model][parameter].value = median[column]
+        for i, (_, source) in enumerate(self.sources):
+            bands[i] = np.vstack(
+                [bands[i][0],
+                 getattr(source, method)(*args), bands[i][1]])
+        return X, bands
+
 
 class Source:
     def __init__(self,
